/*
 * Copyright 2010 Nathan Phillip Brink.
 *
 * bin_replace_string is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * bin_replace_string is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with bin_replace_string.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <fcntl.h>
#include <libelf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

int bin_replace_string(FILE *in, FILE *out, const char *str, const char *rpl);
static int elf_replace_string(int fd, const char *str, const char *rpl);

static size_t bin_replace_string_buf(char *buf, size_t buf_len, const char *str, size_t str_len, const char *rpl, size_t rpl_len, int *dirty);

static int copy(int infd, int outfd);

static int bin_replace_string_usage(FILE *out, const char *progname);

int main(int argc, char *argv[])
{
  FILE *in;
  char *in_filename;
  int infd;

  FILE *out;
  char *out_filename;
  int outfd;

  char *string;
  char *replacement;

  int use_elf;

  int ret;

  int opt;

  in = NULL;
  in_filename = NULL;

  out = NULL;
  out_filename = NULL;

  string = NULL;
  replacement = NULL;

  use_elf = 1;

  ret = 0;

  while ( (opt = getopt(argc, argv, "ho:rv")) != -1 )
    {
      switch (opt)
	{
	case 'h':
	  return bin_replace_string_usage(stdout, argv[0]);

	case 'v':
	  printf(PACKAGE_STRING "\n"
		 PACKAGE_URL "\n"
		 "Copyright 2010 Nathan Phillip Brink\n"
		 "License AGPLv3+: Affero GNU GPL version 3 or later.");
	  return 0;

	case 'o':
	  if (out_filename)
	    {
	      fprintf(stderr, "The output file (option 'o') was already specified. This argument may only be specified once.\n");
	      return 2;
	    }
	  out_filename = optarg;
	  break;

	case 'r':
	  if (!use_elf)
	    {
	      fprintf(stderr, "You specified the 'r' option multiple times.\n");
	    }

	  use_elf = 0;
	  break;
	}
    }

  if (argc - 3 != optind)
    {
      if (argc - 3 < optind)
	fprintf(stderr, "Not enough arguments.\n");
      else
	fprintf(stderr, "Too many arguments.\n");
      fprintf(stderr, "\n");

      bin_replace_string_usage(stderr, argv[0]);
      return 2;
    }

  string = argv[optind ++];
  replacement = argv[optind ++];
  in_filename = argv[optind ++];

  /*
   * The spaghetti-code method of choosing between elf and raw support
   * :-).
   */
  if (use_elf)
    {
      if (out_filename)
	{
	  infd = open(in_filename, O_RDONLY);
	  if (infd < 0)
	    {
	      perror("open");
	      fprintf(stderr, "Unable to open ``%s'' for reading.\n", in_filename);
	      return 1;
	    }

	  outfd = open(out_filename, O_RDWR | O_CREAT, 0700);
	  if (outfd < 0)
	    {
	      perror("open");
	      fprintf(stderr, "Unable to open ``%s'' for writing/update.\n",
		      out_filename);
	      return  1;
	    }

	  if (copy(outfd, infd))
	    {
	      fprintf(stderr, "Error creating copying ``%s'' to ``%s''.\n",
		      in_filename, out_filename);
	      unlink(out_filename);
	      return 1;
	    }

	  if (lseek(outfd, 0, SEEK_SET) == (off_t)-1)
	    {
	      fprintf(stderr, "Error priming ``%s''.\n",
		      out_filename);
	      unlink(out_filename);
	      return 1;
	    }

	  close(infd);
	}
      else
	{
	  outfd = open(in_filename, O_RDWR);
	  if (outfd < 0)
	    {
	      fprintf(stderr, "Error opening ``%s'' for updating.\n",
		      in_filename);
	      return 1;
	    }
	}

      ret = elf_replace_string(outfd, string, replacement);
      close(outfd);

      return ret;
    }

  if (out_filename)
    {
      out = fopen(out_filename, "wb");
      if (!out)
	{
	  perror("fopen");
	  fprintf(stderr, "Error opening ``%s'' for writing.\n", out_filename);
	  return 1;
	}
    }

  if (out)
    in = fopen(in_filename, "rb");
  else
    in = fopen(in_filename, "r+b");
  if (!in)
    {
      perror("fopen");
      fprintf(stderr, "Error opening ``%s'' for %s.\n",
	      in_filename, out ? "reading" : "updating");
      return 1;
    }

  ret = bin_replace_string(in, out, string, replacement);

  fclose(in);
  if (out)
    fclose(out);

  return ret;
}

/**
 * \brief
 *   Display the help/usage to the user.
 *
 * \param out
 *   Stream to use: normally stderr or stdout.
 * \param progname
 *   The name of this executable.
 */
static int bin_replace_string_usage(FILE *out, const char *progname)
{
  fprintf(stderr, "%s [-o <output filename>] [-r] [--] <string to replace> <replacement string> <input filename>\n"
	  "Replace a NULL-terminated string in an executable with another string, padding it with NULLs.\n"
	  "\n"
	  " -h Display this help.\n"
	  " -v Display the package name and version.\n"
	  "\n"
	  " -o <output filename> Write output to this file instead of updating the input file in-place.\n"
	  " -r Support raw files instead interpreting them as ELFs. (Use this if you have to change a string outside of an ELF's .rodata section... but recall that such is evil).\n",
	  progname);

  return 0;
}

/**
 * \brief
 *   Copy data from one file to another.
 *
 * \param dstfd
 *   File to write output to.
 * \param srcfd
 *   File to read input from.
 */
static int copy(int dstfd, int srcfd)
{
  char buf[8192];
  ssize_t inlen;
  ssize_t outlen;
  size_t out_i;

  while ( (inlen = read(srcfd, buf, sizeof(buf))) > 0 )
    {
      out_i = 0;
      while (out_i < inlen)
	{
	  outlen = write(dstfd, buf + out_i, inlen - out_i);
	  if (outlen == -1)
	    {
	      perror("write");
	      fprintf(stderr, "Error copying file to output file: error writing to new file.\n");
	      return 1;
	    }

	  out_i += outlen;
	}
    }
  if (inlen == -1)
    {
      perror("read");
      fprintf(stderr, "Error copying file to output file: error reading original file.\n");
      return 1;
    }

  return 0;
}

/**
 * \brief
 *   Some setup code common to both elf_replace_string() and
 *   bin_replace_string().
 *
 * Will, for example, confirm that strlen(str) >= strlen(rpl).
 *
 * \param str
 *   The string to be replaced.
 * \param str_len
 *   The length of str shall be stored here.
 * \param rpl
 *   The string which replaces str.
 * \param rpl_len
 *   The length of rpl shall be stored here.
 * \return
 *   0 on success, 1 on error.
 */
static int _bin_replace_string_common(const char *str, size_t *str_len,
				      const char *rpl, size_t *rpl_len)
{
  *str_len = strlen(str);
  *rpl_len = strlen(rpl);

  if (*rpl_len > *str_len)
    {
      fprintf(stderr, "Replacement string is longer than the original string. I cannot do that sort of thing.\n");
      return 1;
    }

  return 0;
}

/**
 * \brief
 *   Replace a NULL-terminated string in an ELF's .rodata with another
 *   string.
 *
 * Our interface is restricted by libelf's inability to write out an
 * elf file to a different destination than it is being read from.
 *
 * \param fd
 *   The file descriptor of the file to update. (Must be able to be
 *   both read and written to).
 * \param str
 *   The string which should be replaced.
 * \param rpl
 *   The replacement for str. Must have a shorter length than str.
 * \return
 *   0 on success and 1 on failure.
 */
static int elf_replace_string(int fd, const char *str, const char *rpl)
{
  size_t str_len;
  size_t rpl_len;

  Elf *elf;
  Elf_Scn *elf_section;
  Elf_Data *elf_data;
  Elf_Data *elf_data_writeback;

  char *buf;
  char *newbuf;
  size_t buf_len;
  size_t buf_fill;

  size_t section_stringtable_index;

  size_t elf_rodata_section_index;

  size_t i;
  size_t eaten_bytes;
  int dirty;
  int is_dirty;

  size_t elf_data_writeback_written_bytes;

  if (_bin_replace_string_common(str, &str_len, rpl, &rpl_len))
    return 1;

  if (elf_version(EV_CURRENT) == EV_NONE)
    {
      fprintf(stderr, "libelf refuses to work because I am using headers for a newer version of libelf than is installed.\n");
      return 1;
    }

  elf = elf_begin(fd, ELF_C_RDWR, NULL);
  if (!elf)
    {
      fprintf(stderr, "Failed to load ELF object: %s\n",
	      elf_errmsg(elf_errno()));
      return 1;
    }

  /* find the ELF section string table map */
  if (elf_getshdrstrndx(elf, &section_stringtable_index))
    {
      fprintf(stderr, "Failed to find section string table map: %s\n",
	      elf_errmsg(elf_errno()));
      return 1;
    }

  elf_section = elf_getscn(elf, section_stringtable_index);
  if (!elf_section)
    {
      fprintf(stderr, "Unable to load section %lu: %s\n",
	      section_stringtable_index, elf_errmsg(elf_errno()));
      return 1;
    }

  buf_len = 0;
  i = 0;
  elf_data = NULL;
  elf_rodata_section_index = 0;
  while ( (elf_data = elf_getdata(elf_section, elf_data)) )
    {
      buf = elf_data->d_buf;
      buf_len = elf_data->d_size;

      for (i = 0;
	   i < buf_len && strncmp(".rodata", &buf[i], buf_len - i);
	   i += strnlen(&buf[i], buf_len - i) + 1)
	{
	  elf_rodata_section_index ++;
	}
      if (i < buf_len)
	break;
    }
  buf = NULL;

  /* detect failure to find .rodata */
  if (i >= buf_len)
    {
      fprintf(stderr, "Unable to find section ``.rodata'' in ELF object.\n");
      return 1;
    }
  buf_len = 0;

  elf_section = elf_getscn(elf, elf_rodata_section_index);
  if (!elf_section)
    {
      fprintf(stderr, "Unable to get section %lu: %s",
	      elf_rodata_section_index, elf_errmsg(elf_errno()));
      return 1;
    }

  buf_len = 2048;
  buf_fill = 0;
  buf = malloc(buf_len);
  if (!buf)
    {
      fprintf(stderr, "Unable to allocate initial buffer\n");
      return 1;
    }

  elf_data = NULL;
  elf_data_writeback = elf_getdata(elf_section, NULL);
  if (!elf_data_writeback)
    {
      fprintf(stderr, "Error initializing writeback buffers.\n");
      return 1;
    }

  elf_data_writeback_written_bytes = 0;
  is_dirty = 0;
  while ( (elf_data = elf_getdata(elf_section, elf_data)) )
    {
      if (buf_len < buf_fill + elf_data->d_size)
	{
	  buf_len = buf_fill + elf_data->d_size;
	  newbuf = malloc(buf_len);
	  if (!newbuf)
	    {
	      fprintf(stderr, "Unable to expand buffer\n");
	      return 1;
	    }
	  memcpy(newbuf, buf, buf_fill);

	  free(buf);
	  buf = newbuf;
	}

      memcpy(&buf[buf_fill], elf_data->d_buf, elf_data->d_size);
      buf_fill += elf_data->d_size;

      eaten_bytes = bin_replace_string_buf(buf, buf_fill, str, str_len, rpl, rpl_len, &dirty);
      if (dirty)
	is_dirty = 1;

      while (eaten_bytes)
	{
	  while (elf_data_writeback->d_size < elf_data_writeback_written_bytes + eaten_bytes)
	    {
	      memcpy(((char *)elf_data_writeback->d_buf) + elf_data_writeback_written_bytes, buf, elf_data_writeback->d_size - elf_data_writeback_written_bytes);

	      eaten_bytes -= elf_data_writeback->d_size - elf_data_writeback_written_bytes;
	      buf_fill -= elf_data_writeback->d_size - elf_data_writeback_written_bytes;
	      memmove(buf, &buf[elf_data_writeback->d_size - elf_data_writeback_written_bytes], buf_fill);

	      elf_data_writeback = elf_getdata(elf_section, elf_data_writeback);
	      if (!elf_data_writeback)
		{
		  fprintf(stderr, "Unable to write back anymore.\n");
		  return 1;
		}
	      elf_data_writeback_written_bytes = 0;
	    }

	  memcpy(((char *)elf_data_writeback->d_buf) + elf_data_writeback_written_bytes, buf, eaten_bytes);

	  elf_data_writeback_written_bytes += eaten_bytes;
	  buf_fill -= eaten_bytes;
	  eaten_bytes = 0;
	  memmove(buf, &buf[eaten_bytes], buf_fill);
	}
    }
  /*
   * we won't close off the last data section because a few bytes will
   * remain unprocessed because we only process them if the number of
   * bytes >= str_len. we may not even write to it if str_len is very,
   * very big
   */

  /* we're done playing with buffers now :-) */
  free(buf);

  if (is_dirty)
    {
      /*
       * We are too lazy to mark individual data sections as
       * dirty. This will make sure that the whole section gets
       * rewritten.
       */
      elf_flagscn(elf_section, ELF_C_SET, ELF_F_DIRTY);

      /*
       * Don't allow libelf to rearrange the sections when it writes
       * out the file. When it's given the liberty, it changes random
       * thing and breaks executables just because it thinks doing so
       * is fun.
       */
      elf_flagelf(elf, ELF_C_SET, ELF_F_LAYOUT);

      if (elf_update(elf, ELF_C_WRITE) <= 0)
	{
	  fprintf(stderr, "Failed to save changes to ELF object: %s\n",
		  elf_errmsg(elf_errno()));
	  return 1;
	}
    }

  elf_end(elf);

  return 0;
}

/**
 * \brief
 *   Replace a NULL-terminated string from the input and write to output.
 *
 * \param in
 *   The stream containing text to replace.
 * \param out
 *   The stream to write out the changed text or NULL if the text
 *   should be written out to the same as the input using seeking.
 * \param str
 *   The string which should be replaced.
 * \param rpl
 *   The replacement for str. Must have a shorter length than str.
 */
int bin_replace_string(FILE *in, FILE *out, const char *str, const char *rpl)
{
  char *buf;

  size_t str_len;
  size_t rpl_len;

  size_t buf_len;

  size_t eaten_bytes;
  size_t written_bytes;
  int dirty;

  FILE *inout;

  if (_bin_replace_string_common(str, &str_len, rpl, &rpl_len))
    return 1;

  if (out)
    inout = out;
  else
    inout = in;

  buf = malloc(str_len * 2);
  buf_len = fread(buf, 1, (str_len + 1) * 2, in);

  while (buf_len >= str_len)
    {
      eaten_bytes = bin_replace_string_buf(buf, buf_len, str, str_len, rpl, rpl_len, &dirty);
      if (!out && dirty)
	if (fseek(inout, -buf_len, SEEK_CUR))
	  {
	    perror("fseek");
	    fprintf(stderr, "You have asked me to edit a file in place, but I have an unseekable stream\n");
	    free(buf);
	    return 1;
	  }

      if (out || dirty)
	{
	  written_bytes = fwrite(buf, 1, eaten_bytes, inout);
	  if (written_bytes != eaten_bytes)
	    {
	      perror("fwrite");
	      fprintf(stderr, "Error writing out data.\n");
	      return 1;
	    }
	}

      if (!out && dirty)
	if (fseek(inout, buf_len, SEEK_CUR))
	  {
	    perror("fseek");
	    fprintf(stderr, "Unable to seek forward to original position and read more data\n");
	    free(buf);
	    return 1;
	  }

      buf_len -= eaten_bytes;
      memmove(buf, &buf[eaten_bytes], buf_len);

      buf_len += fread(&buf[buf_len], 1, (str_len + 1) * 2 - buf_len, in);
    }

  if (out)
    {
      written_bytes = fwrite(buf, 1, buf_len, out);
      if (written_bytes != buf_len)
	{
	  perror("fwrite");
	  fprintf(stderr, "Error writing out data.\n");
	  return 1;
	}
    }

  free(buf);

  return 0;
}

/**
 * \brief
 *   Replace all occurances of a NULL-terminated string in a buffer.
 *
 * This performs the actual string replacement on a given buffer,
 * making up the only common code between the whole-binary text
 * replacement and ELF support.
 *
 * If data is being read in on an as-needed basis, the buffer will
 * likely not hold the entire amount of data on which string
 * replacement will be performed. In this case, data will be given in
 * chunks. Mind that each chunk must be at least as long as
 * strlen(str). Also, mind that the return value is especially
 * important in this situation as it specifies how much of buf has
 * been edited and may be accepted as final whereas the remainder of
 * buf must be kept around for the next round. This will vary
 * depending on whether or not a string replacement has taken place.
 *
 * If you are feeding data gradually through this function, you must
 * know that there can be nothing to replace when buf_len <
 * strlen(str). You can either detect this yourself or detect this by
 * noting that the return value will be 0. Make sure that you provide
 * enough data to this function so that this return value of 0 isn't
 * triggered when there is still more data to read.
 *
 * \param buf
 *   The buffer which should be edited.
 * \param buf_len
 *   The length of the input buffer.
 * \param str
 *   The string to be replaced.
 * \param str_len
 *   The return of strlen(str) so that I don't have to calculate that
 *   each time myself.
 * \param rpl
 *   The string to replace str.
 * \param rpl_len
 *   The return of strlen(rpl) so that I don't have to calculate that
 *   each time myself.
 * \param dirty
 *   If not NULL, this will be set to 1 if buf was modified and 0 if
 *   buf wasn't modified. This is helpful to figure out if fseek()
 *   needs to be called, for example.
 * \return
 *   The number of bytes which need not be revisited again. This will
 *   generally be buf_len - strlen(str). If buf_len is too short for
 *   any replacement to happen (i.e., it's shorter than the length of
 *   str), then the return value will be 0. Ensure that you're feeding
 *   an adequately-sized buffer to this function!
 */
static size_t bin_replace_string_buf(char *buf, size_t buf_len, const char *str, size_t str_len, const char *rpl, size_t rpl_len, int *dirty)
{
  size_t i;

  if (dirty)
    *dirty = 0;

  for (i = 0; i + str_len <= buf_len;)
    {
      if (!memcmp(&buf[i], str, str_len + 1))
	{
	  if (dirty)
	    *dirty = 1;

	  memcpy(&buf[i], rpl, rpl_len);
	  i += rpl_len;

	  /* NULL-terminate and pad */
	  memset(&buf[i], 0, str_len - rpl_len);
	  i += str_len - rpl_len;
	}
      else
	{
	  i ++;
	}
    }

  return i;
}
